const vm = require('vm');

// filter out 'GLOBAL' and 'root' as these generate deprecation warnings
const safeContext = `
(function() {
	Function = undefined;
	const keys = Object.getOwnPropertyNames(this)
		.filter(a => a !== 'GLOBAL' && a !== 'root')
		.concat(['constructor']);
	for (const key of keys) {
		const item = this[key];
		if (item && typeof item.constructor === 'function') {
			this[key].constructor = undefined;
		}
	}
})();
`;

function execute(req, cb) {
	const { data, code, unsafe } = req.params;
	if (typeof code !== 'string') {
		return cb.error(null, 'invalid argument: code');
	}
	// determine the correct action to use
	const action = unsafe ? executeUnsafe : executeSafe;

	action(data, code).then(result => {
		cb.next(null, result);
	}, ex => {
		cb.error(null, formatException(ex));
	});
}

async function executeSafe(data, code) {
	// Define a new async function to resolve within a sandbox environment
	const script = new vm.Script(`(async function() { ${safeContext} ${code} })();`);
	return script.runInNewContext({ data });
}

async function executeUnsafe(data, code) {
	// Define a new async function to resolve without a sandbox environment
	// eslint-disable-next-line no-new-func
	const fn = new Function('data', 'require', `return (async function() { ${code} })();`);
	// to be functionally consistent with the safe version, make 'this' the global object
	// A number of items are not global - require, export, module, __dirname, __filename.
	// We just care about being able to use require in this function during unsafe mode.
	return fn.call(global, data, require);
}

function formatException(ex) {
	// The code provided by the user may not throw/reject with an Error object.
	// We would check ex instanceof Error but the different context of the vm causes this to fail
	if (ex instanceof Error || (typeof ex === 'object' && ex.stack && ex.message)) {
		// return a new object with the error message as an explicit property
		// JS Errors do not show the message when stringified.
		return {
			message: ex.message
		};
	}
	return ex;
}

module.exports = execute;
